import networkx as nx
import statistics
from datetime import date
import shapely.affinity as sa
import shapely.ops as so
from shapely.prepared import prep
import os
from pathlib import Path
from utilities.dcmgeometrysdk.geometryfunctions.transformationfunctions import transform_geoms, transform_coordinates
from utilities.dcmgeometrysdk.geometryfunctions.transformationfunctions import helmert_transformation_with_ids
from utilities.dcmgeometrysdk.geometryfunctions.bearingdistancefunctions import *
from utilities.dcmgeometrysdk.geometryfunctions.conversionsfunctions import dd2hp
from utilities.dcmgeometrysdk.geometryfunctions.otherfunctions import previous_and_next, chunker
from utilities.dcmgeometrysdk.dcmgeometry.factories.polygonfactory import PolygonGeomFactory
from utilities.dcmgeometrysdk.dcmgeometry.surveygraph import SurveyGraph
from utilities.dcmgeometrysdk.dcmgeometry.factories.arclinefactory import ArcLineGeomFactory
from utilities.dcmgeometrysdk.dcmgeometry.factories.pointfactory import PointGeomFactory
from utilities.dcmgeometrysdk.dcmgeometry.lines import LineGeom
from utilities.dcmgeometrysdk.dcmgeometry.admin import Admin
from utilities.dcmgeometrysdk.dcmgeometry.loops import Loops
from utilities.dcmgeometrysdk.dcmgeometry.polygons import PolygonGeom
from utilities.dcmgeometrysdk.landxml.landxml import LandXML, CgPoint, InstrumentSetup, \
                                                        InstrumentPoint, ReducedObservation, ParcelType
from utilities.dcmgeometrysdk.geometryfunctions.misclosefunctions import loop_checker


class Geometries:
    def __init__(self, landxml=None, mis_tol=.1):
        self.landxml = landxml
        self.points = None
        self.lines = None
        self.polygons = None
        self.loops = None
        self.dataframes = {}
        self.target_points = None
        self.crs = None
        self.swing_value = 0
        self.transform_crs = None
        self.target_points = None
        self.mis_tol = mis_tol
        self.survey_graph = None
        self.survey_number = None
        self.survey_year = None
        self.admin = None
        self.ccc = None

        if landxml is not None:
            landxml = self.add_plan_features_to_parcels(landxml)
            is2point = self.get_is_2_point()
            self.add_missing_points_lines()
            self.points = PointGeomFactory().build(self.landxml, is2point)
            self.crs = self.set_crs()
            self.lines = ArcLineGeomFactory().build(self.landxml, self.points, is2point, self.crs)
            self.survey_graph = SurveyGraph(self.lines, self.points)
            self.polygons = PolygonGeomFactory().build(self.landxml, self.lines, self.points, self.crs)
            self.survey_number = self.get_survey_number()
            self.survey_year = self.get_survey_year()
            self.admin = Admin(self.landxml.Survey[0].SurveyHeader, self.polygons, self.landxml.CoordinateSystem)
            self.ccc = self.set_ccc()
            self.loops = self.set_loop_errors()
            self.original_crs = self.set_original_crs()

    def add_plan_features_to_parcels(self, landxml):
        for item in landxml.PlanFeatures:
            name = item.name
            pf_desc = item.desc
            for pf in item.PlanFeature:
                polygon = ParcelType()
                polygon.name = pf.name
                polygon.class_ = name
                polygon.parcelFormat = 'Geometry'
                polygon.state = 'Occupation'
                polygon.Center = []
                if isinstance(pf.CoordGeom, list):
                    polygon.CoordGeom = pf.CoordGeom
                else:
                    polygon.CoordGeom = [pf.CoordGeom]
                polygon.desc = pf_desc
                if pf.desc is not None:
                    if len(pf.desc.strip()) > 0:
                        polygon.desc = pf.desc
                polygon.parcelType = "Single"
                landxml.Parcels[0].Parcel.append(polygon)
        return landxml
    """the following 2 methods link the lines to the points, and the lines to the polygons, vice versa"""

    # instrument setup id to point id reference dictionary
    def get_is_2_point(self):
        if self.landxml is not None:
            return {i.id: i.InstrumentPoint[0].pntRef for i in self.landxml.Survey[0].InstrumentSetup}

    # point id to instrument setup id reference dictionary
    def get_point_2_is(self):
        is2point = self.get_is_2_point()
        if is2point is not None:
            return {v: k for k, v in is2point}

    def add_irregular_lines(self, temp_lines):
        for k, v in temp_lines.items():
            if k not in self.lines:
                self.lines[k] = v

    def get_survey_number(self):
        survey_header = self.landxml.Survey[0].SurveyHeader
        survey_number = survey_header.name
        if survey_number is None:
            survey_number = ''
        return survey_number

    def set_loop_errors(self, mis_tol=None):
        if mis_tol is None:
            mis_tol = self.mis_tol
        if mis_tol is None:
            mis_tol = .1

        loops = loop_checker(self, self.ccc, mis_tol=mis_tol)
        f_loops = {}
        all_likely = []
        for k, v in loops.items():
            f_loops[k] = Loops(k, v, self.lines)
            all_likely.extend(v.get('likely', []))
        all_likely = set(all_likely)
        for k, v in self.lines.items():
            if v.name in all_likely:
                v.likely_candidate = True

        return f_loops

    def set_ccc(self):
        d = dict(self.survey_graph.graph.degree())
        id_name = dict()
        for key, value in d.items():
            id_name.setdefault(value, list()).append(key)
        ccc = (sorted(id_name[max(id_name)]))[0]
        self.points.get(ccc).ccc = True
        return ccc

    def get_survey_year(self):
        survey_header = self.landxml.Survey[0].SurveyHeader
        dates = survey_header.AdministrativeDate
        year = 1900

        if self.survey_number.startswith('TP'):
            year = 1900
        else:
            if not isinstance(dates, list):
                dates = [dates]
            for dated in dates:
                if dated.adminDateType == 'Date of Survey':
                    survey_date = dated.adminDate
                    if survey_date is None:
                        survey_date = date(year=1900, month=1, day=1)
                    try:
                        year = survey_date.year
                    except ValueError:
                        year = 1900
        return year

    def estimate_swing(self):
        swings = []
        for key, value in self.lines.items():
            calcd = math.degrees(calc_bearing(self.points.get(key[1]).geometry,
                                              self.points.get(key[0]).geometry))
            dif = calcd - value.dd_bearing
            if dif < -180:
                dif += 360
            swings.append(dif)

        if len(swings) > 0:
            std = statistics.pstdev(swings)
            mean = statistics.mean(swings)
            swings = [x for x in swings if (x > mean - 2 * std) and (x < mean + 2 * std)]
            self.swing_value = statistics.mean(swings)
        else:
            self.swing_value = 0

    def recalc_geometries(self, ref_point=None, swing=False, sf=1, swing_value=None, xml=True):
        self.swing_value = swing_value
        if swing is True and swing_value is None:
            self.estimate_swing()
        else:
            self.swing_value = 0

        if self.points.get(ref_point) is None:
            connected = nx.is_k_edge_connected(self.survey_graph.graph, 1)
            if connected is False:
                branches = list(nx.k_edge_components(self.survey_graph.graph, 1))
                max_b = max(branches, key=len)
                ref_point = sorted([item for item in max_b])[-1]
            else:
                ref_points = sorted(list(self.survey_graph.graph.nodes))
                ref_point = ref_points[0]

        short_paths = nx.single_source_shortest_path(self.survey_graph.graph, ref_point)
        ref_point_coords = self.points.get(ref_point).geometry.coords[:]

        new_geom = set()
        for key, value in short_paths.items():
            if self.points.get(ref_point).geometry.has_z is False:
                ref_e, ref_n = ref_point_coords[0]
            else:
                ref_e, ref_n, ref_z = ref_point_coords[0]

            for prev, item, nxt in previous_and_next(value):
                if nxt is not None:
                    b_d = self.survey_graph.graph.get_edge_data(item, nxt)
                    bearing = b_d['bearing']
                    if (item, nxt) != b_d['st']:
                        bearing += 180
                        bearing = calc_inside_360(bearing)

                    delta_e, delta_n = calc_new_point(bearing, b_d['distance'] * sf)
                    ref_e += delta_e
                    ref_n += delta_n

                    if nxt not in new_geom:
                        new_geom.add(nxt)
                        self.points[nxt].geometry = sg.Point(ref_e, ref_n)

        if swing is True and self.swing_value != 0:
            self.apply_swing(self.swing_value, origin=self.points.get(ref_point).geometry)
        else:
            self.update_geometries(self.swing_value, xml=xml)

    def transform_geometries(self, out_proj, use_xml_data=True):
        self.points = transform_geoms(self.points, self.crs, out_proj)

        if use_xml_data is True and self.landxml is not None:
            self.update_geometries()
        else:
            if self.lines is not None:
                self.lines = transform_geoms(self.lines, self.crs, out_proj)
            if self.polygons is not None:
                self.polygons = transform_geoms(self.polygons, self.crs, out_proj)
                for k, v in self.polygons.items():
                    v.set_coord_lookup(self.points)
            if self.loops is not None:
                if len(self.loops) > 0:
                    self.loops = transform_geoms(self.loops, self.crs, out_proj)
            if self.admin is not None:
                if self.admin.geometry is not None:
                    ag = transform_geoms({'i': self.admin}, self.crs, out_proj)
                    self.admin = ag['i']
        self.crs = out_proj

    def reset_geometries_to_original_crs(self):
        self.transform_geometries(out_proj=self.original_crs)

    # just take crs of first point
    # assuming only 1 crs for entire plan, this never changes in Aus.
    def set_crs(self):
        for p in self.points.values():
            crs = p.crs
            break
        return crs

    def set_original_crs(self):
        for p in self.points.values():
            crs = p.original_crs
            break
        return crs

    def set_target_points(self, target_points):
        self.target_points = target_points

    def transform_onto_points_with_ids(self, target_points=None):
        # target points is a dictionary of points
        # {point_id: sg.Point()}
        if target_points is None:
            target_points = self.target_points
        self.points = helmert_transformation_with_ids(self.points, target_points, self.lines)
        return self.points

    def add_missing_points_lines(self):
        existing_points = set(point.valueOf_.strip() for point in self.landxml.CgPoints[0].get_CgPoint())
        lxml_parcels = self.landxml.Parcels
        lxml_polygons = []
        if len(lxml_parcels) > 0:
            for x in lxml_parcels:
                lxml_polygons += x.Parcel
        for polygon in lxml_polygons:
            if polygon.parcelType != 'Multipart':
                if (polygon.class_ == 'Easement' and polygon.parcelFormat == 'Standard') is False:
                    if len(polygon.CoordGeom) > 0:
                        coord_geom = polygon.CoordGeom[0]
                        irls = coord_geom.IrregularLine
                        ir_count = 0
                        c = 90000
                        for irregular_line in irls:
                            sp = irregular_line.Start.pntRef
                            tp = irregular_line.End.pntRef
                            for p in [sp, tp]:
                                insetup = InstrumentSetup()
                                insetup.id = p.replace('CGPNT', 'IS')
                                insetup.stationName = insetup.id
                                insetup.instrumentHeight = 0
                                insetup.add_InstrumentPoint(value=InstrumentPoint(p))
                                self.landxml.Survey[0].add_InstrumentSetup(value=insetup)
                            ro = ReducedObservation()
                            default = c + ir_count
                            ro.name = f'OBS-{default}'
                            ro.desc = 'Irregular Line'
                            ro.azimuthType = 'computed'
                            ro.distanceType = 'computed'
                            ro.purpose = 'irregular line'
                            ro.setupID = sp.replace('CGPNT', 'IS')
                            ro.targetSetupID = tp.replace('CGPNT', 'IS')

                            pnt_list = irregular_line.PntList2D
                            geom_value = []
                            chunk_list = None
                            if pnt_list is not None:
                                chunk_list = [i for i in chunker(irregular_line.PntList2D.split(), 2)]
                                for n, e in chunk_list:
                                    if f'{n} {e}' not in existing_points:
                                        point = CgPoint()
                                        point.valueOf_ = f'{n} {e}'
                                        point.name = f'IR-{n}-{e}'
                                        point.pntSurv = 'natural boundary'
                                        point.state = 'existing'

                                        self.landxml.CgPoints[0].add_CgPoint(value=point)

                                        geom_value.append((float(e),float(n)))

                            else:
                                pnt_list = irregular_line.PntList3D
                                if pnt_list is not None:
                                    chunk_list = [i for i in chunker(irregular_line.PntList2D.split(), 3)]
                                    for n, e, z in chunk_list:
                                        if f'{n} {e}' not in existing_points:
                                            point = CgPoint()
                                            point.valueOf_ = f'{n} {e} {z}'
                                            point.name = f'IR-{n}-{e}-{z}'
                                            point.pntSurv = 'natural boundary'
                                            point.state = 'existing'
                                            self.landxml.CgPoints[0].add_CgPoint(value=point)
                                            geom_value.append((float(e), float(n), float(z)))

                            if chunk_list is not None:
                                sgeom = sg.Point(float(chunk_list[0][1]), float(chunk_list[0][0]))
                                egeom = sg.Point(float(chunk_list[-1][1]), float(chunk_list[-1][0]))
                                ro.horizDistance = calc_distance(sgeom, egeom)
                                ro.azimuth = dd2hp(calc_bearing(egeom, sgeom))
                                self.landxml.Survey[0].ObservationGroup[0].add_ReducedObservation(value=ro)
                            ir_count += 1

    def apply_translation(self, x=0, y=0):
        if x != 0 and y != 0:
            for k, v in self.points.items():
                self.points[k].geometry = sa.translate(v.geometry, x, y)
            self.update_geometries()

    def apply_swing(self, angle=0, origin='center'):
        """origin here should be the point you want to swing the plan around.
        defaulted to the centre of the a point so you wont get any rotation of a point feature if its not set,
        this will generally be the point that you translate the survey onto
        angle is in decimal degrees"""
        if angle != 0:
            for k, v in self.points.items():
                self.points[k].geometry = sa.rotate(v.geometry, -1 * angle, origin)
            self.update_geometries(swing=angle)

    def apply_affine_transformation(self, mat=None):
        if mat is not None:
            for k, v in self.points.items():
                self.points[k].geometry = sa.affine_transform(v.geometry, mat)
            self.update_geometries()

    def update_geometries(self, swing=0, xml=True):
        if xml is True:
            self.lines = ArcLineGeomFactory().build(self.landxml, self.points, self.get_is_2_point(), self.crs, swing)
            self.polygons = PolygonGeomFactory().build(self.landxml, self.lines, self.points, self.crs)
            self.admin.geometry = self.admin.set_geometry(self.polygons)
        else:
            line: LineGeom
            for line in self.lines.values():
                line.setup_point = self.points.get(line.setup_point.name)
                line.target_point = self.points.get(line.target_point.name)
                line.calc_geometry_from_start_end()
        self.survey_graph = SurveyGraph(self.lines, self.points)
        self.loops = self.set_loop_errors()

    def write_geom_to_file(self, points=True, lines=True, arcs=True, polygons=True, location=None, loops=False,
                           file_type='GPKG', same_file=True, df_only=False, filename_suffix='', remove=None):

        def clean_columns(df):
            for column in df.columns:


                if column == 'setup_point' or column == 'target_point':
                    try:
                        df[column] = df[column].apply(lambda x: x.point_oid)
                    except Exception as err:
                        print(err)
            return df

        def drop_columns(df, columns):
            for column in columns:
                if column in df.columns:
                    df.drop(columns=[column], inplace=True)
            return df

        def make_gdf(vals, cols=None):
            if cols is None:
                cols = []

            df = pd.DataFrame(vals)
            df = drop_columns(df, cols)
            df = clean_columns(df)
            gdf = gpd.GeoDataFrame(df.drop(columns='geometry'), geometry=df['geometry'], crs=self.crs)
            gdf['filename'] = self.survey_number
            return gdf

        def write_file(gdf, layer, location, suffix, file_type, same_file):
            out_path = Path(location, self.survey_number)
            out_path.mkdir(exist_ok=True, parents=True)
            if same_file is False or file_type != 'GPKG':
                out_path = Path(out_path, layer + filename_suffix + suffix)
                gdf.to_file(str(out_path), driver=file_type)
            else:
                out_path = Path(out_path, self.survey_number + filename_suffix + suffix)
                gdf.to_file(str(out_path), driver=file_type, layer=layer)

        extension_lookup = {'GPKG': '.gpkg', 'Esri ShapeFile': '.shp'}

        try:
            import pandas as pd
            import geopandas as gpd
        except ImportError as e:
            print(e)
            raise
        if file_type not in {'GPKG', 'Esri Shapefile'}:
            file_type = 'Esri Shapefile'
            print('File type wasnt recognised, setting to shapefile')

        suffix = extension_lookup.get(file_type)

        if location is None and df_only is False:
            print('Location cannot be None')
            raise NotADirectoryError
        elif location is not None:
            if os.path.exists(location) is False and df_only is False:
                out_path = Path(location)
                out_path.mkdir(exist_ok=True, parents=True)
                # print('Location must exist')
                # raise NotADirectoryError

        if points is True:
            vals = [v.__dict__ for v in self.points.values()]
            if remove is None:
                remove = ['neh', 'original_geom']
            gdf = make_gdf(vals, remove)
            self.dataframes['points'] = gdf
            if df_only is False:
                write_file(gdf, 'points', location, suffix, file_type, same_file)

        if loops is True:
            vals = [v.__dict__ for v in self.loops.values()]
            gdf = make_gdf(vals)
            self.dataframes['loops'] = gdf
            if df_only is False:
                write_file(gdf, 'loops', location, suffix, file_type, same_file)

        if lines is True or arcs is True:
            lines_vals = []
            arcs_vals = []

            for v in self.lines.values():
                if isinstance(v, LineGeom):
                    lines_vals.append(v.__dict__)
                else:
                    arcs_vals.append(v.__dict__)

            if len(lines_vals) > 0 and lines is True:
                gdf = make_gdf(lines_vals)
                self.dataframes['lines'] = gdf
                if df_only is False:
                    write_file(gdf, 'lines', location, suffix, file_type, same_file)
            if len(arcs_vals) > 0 and arcs is True:
                gdf = make_gdf(arcs_vals)
                self.dataframes['arcs'] = gdf
                if df_only is False:
                    write_file(gdf, 'arcs', location, suffix, file_type, same_file)

        if polygons is True:

            vals = [v.__dict__ for v in self.polygons.values()]
            if remove is None:
                remove = []
            gdf = make_gdf(vals, ['line_order', 'polygon_points', 'inner_angles', 'coord_lookup',
                                  'point_lookup', 'original_geom', 'misclose', 'parcel_arcs', 'arc_rot_errors',
                                  'arc_rad_errors', 'children', ] + remove)

            gdf['polygon_notations'] = gdf['polygon_notations'].apply(lambda x: ','.join(map(str, x)))

            self.dataframes['polygons'] = gdf
            if df_only is False:
                write_file(gdf, 'polygons', location, suffix, file_type, same_file)

    def set_likely_candiates(self):
        for k, v in self.loops:
            for name in v.likely_names:
                line = self.lines.get(name)
                if line is None:
                    line = self.lines.get(line[1], line[0])
                if line is not None:
                    line.likely_candidate = True

    def set_overlapping_lines(self):
        lines_nn = {k: v for k, v in self.lines.items() if v.line_type == 'normal'}
        geoms = []
        rev_lookup = {}
        for k, v in lines_nn.items():
            geoms.append(v.geometry)
            rev_lookup[v.geometry.wkb] = k

        for item, line in lines_nn.items():
            line_geom = line.geometry
            geoms.remove(line_geom)
            if len(geoms) > 0:
                prepped_geom = prep(line_geom)
                hits = list(filter(prepped_geom.crosses, geoms))
                if len(hits) > 0:
                    self.lines[item].overlapping = True
                    for hit in hits:
                        line_hit = rev_lookup[hit.wkb]
                        self.lines[line_hit].overlapping = True

    def gen_graph_from_points_lines(self, add_unconnected=True, add_gen_lines=False):
        if self.points is not None and self.lines is not None:
            self.survey_graph = SurveyGraph(self.lines, self.points, add_unconnected, add_gen_lines)

    def produce_dissolved_polygon(self, types=('created',), include_easements=False):
        if include_easements is True:
            created = so.unary_union([v.geometry for k, v in self.polygons.items() if v.parcel_state in types
                                      and v.geometry is not None])
        else:
            created = so.unary_union([v.geometry for k, v in self.polygons.items()
                                      if v.parcel_state in types and v.easement is False
                                      and v.geometry is not None])

        p = PolygonGeom()
        crs = None
        name = None
        if len(list(self.polygons.values())) > 0:
            crs = list(self.polygons.values())[0].crs
            name = list(self.polygons.values())[0].name
        p.create_polygon(created, self.points, name, crs)
        return p

    def update_xml_coordinates(self):
        points = self.landxml.CgPoints[0].CgPoint
        for point in points:
            name = point.get_name()
            p = self.points.get(name)
            if p is not None:
                coords = p.geometry.coords[:][0]
                n = str(coords[1])
                e = str(coords[0])
                point.latitude = float(n)
                point.longitude = float(e)
                if len(coords) > 2:
                    h = str(coords[2])
                    point.valueOf_ = f'{n} {e} {h}'

                else:
                    point.valueOf_ = f'{n} {e}'

    def write_xml_to_file(self, outfile):
        with open(outfile, 'w') as open_file:
            self.landxml.export(outfile=open_file, level=0)




    # def gen_point_lines_from_graph(self, graph=None):
    #     if graph is None:
    #         graph = self.survey_graph.graph
    #
    #     for









