import networkx as nx
import shapely.ops as so
import shapely.geometry as sg
import math
from utilities.dcmgeometrysdk.dcmgeometry.lines import LineGeom
from utilities.dcmgeometrysdk.geometryfunctions.bearingdistancefunctions import calc_distance, \
    calc_bearing, calc_inside_360



class SurveyGraph:
    def __init__(self, lines, points, add_unconnected=True, add_gen_lines=False):

        self.connected = None
        self.unconnected_nodes = set()
        self.unconnected_branches = set()
        self.graph = self.generate_graph(lines, points, add_unconnected, add_gen_lines)

    def return_branches(self, graph=None):
        if graph is None:
            graph = self.graph
        return list(nx.k_edge_components(graph, 1))

    def generate_graph(self, lines, points, add_unconnected=True, add_gen_lines=False, remove=tuple()):
        g = nx.Graph()
        nodes = set()
        for k, v in lines.items():
            if v.distance_type not in remove:
                setup, target = k
                g.add_edge(setup, target, distance=v.distance, bearing=v.dd_bearing, st=(setup, target),
                           distance_type=v.distance_type, az_type=v.azimuth_type)
                nodes.add(setup)
                nodes.add(target)

        # find unconnected points
        self.unconnected_nodes = set()
        self.unconnected_branches = set()
        for point in points.keys():
            if point not in nodes:
                self.unconnected_nodes.add(point)

        self.connected = nx.is_k_edge_connected(g, 1)
        if self.connected is False:
            # get largest branch
            branches = list(nx.k_edge_components(g, 1))
            max_b = max(branches, key=len)
            ref_point = sorted([item for item in max_b])[-1]
            for item in branches:
                if item != max_b:
                    self.unconnected_nodes.add(list(item)[0])
                    for p in item:
                        self.unconnected_branches.add(p)

        else:
            ref_points = sorted(list(nodes))
            ref_point = ref_points[0]

        # add in generated edges for unconnected points.
        if add_unconnected is True:
            rev_look = {v.geometry.coords[:][0]: k for k, v in points.items()}

            con_multi = sg.MultiPoint([v.geometry for k, v in points.items() if k not in
                                       self.unconnected_nodes.union(self.unconnected_branches)])

            counter = 0
            for point in self.unconnected_nodes:
                target = points.get(point).geometry
                close_point = so.nearest_points(target, con_multi)[1].coords[:][0]
                ref = rev_look.get(close_point, ref_point)
                setup = points.get(ref).geometry
                if setup is not None and target is not None:
                    distance = calc_distance(setup, target)
                    azimuth = calc_bearing(target, setup)
                    azimuth = calc_inside_360(math.degrees(azimuth))
                    g.add_edge(ref, point, distance=distance, bearing=azimuth, st=(ref, point),
                               distance_type='Generated', az_type='Generated')
                    if add_gen_lines is True:

                        line = LineGeom()
                        sp = points.get(ref)
                        tg = points.get(point)
                        line.create_line_from_coords(sp, tg, name=f'generated-{str(counter)}', line_type='Calculated',
                                                     crs=sp.crs)
                        lines[(sp.name, tg.name)] = line
                        counter += 1

        return g

    def ignore_line_type(self, line_types=('Ignored', 'Generated')):
        return nx.Graph(((u, v, e) for u, v, e in self.graph.edges(data=True) if
                        (e['distance_type'] not in line_types and e['az_type'] not in line_types)))
